from jinja2 import Environment, meta
from typing import List, Optional, Union


class QuestionOptionProcessor:
    """
    Class that manages the processing of question options.
    These can be provided directly, as a template string, or fetched from prior answers or the scenario.
    """

    def __init__(self, prompt_constructor):
        self.prompt_constructor = prompt_constructor

    @staticmethod
    def _get_default_options() -> list:
        """Return default placeholder options."""
        return [f"<< Option {i} - Placeholder >>" for i in range(1, 4)]

    @staticmethod
    def _parse_template_variable(template_str: str) -> str:
        """
        Extract the variable name from a template string.

        Args:
            template_str (str): Jinja template string

        Returns:
            str: Name of the first undefined variable in the template

        >>> QuestionOptionProcessor._parse_template_variable("Here are some {{ options }}")
        'options'
        >>> QuestionOptionProcessor._parse_template_variable("Here are some {{ options }} and {{ other }}")
        Traceback (most recent call last):
        ...
        ValueError: Multiple variables found in template string
        >>> QuestionOptionProcessor._parse_template_variable("Here are some")
        Traceback (most recent call last):
        ...
        ValueError: No variables found in template string
        """
        env = Environment()
        parsed_content = env.parse(template_str)
        undeclared_variables = list(meta.find_undeclared_variables(parsed_content))
        if not undeclared_variables:
            raise ValueError("No variables found in template string")
        if len(undeclared_variables) > 1:
            raise ValueError("Multiple variables found in template string")
        return undeclared_variables[0]

    @staticmethod
    def _get_options_from_scenario(
        scenario: dict, option_key: str
    ) -> Union[list, None]:
        """
        Try to get options from scenario data.

        >>> from edsl import Scenario
        >>> scenario = Scenario({"options": ["Option 1", "Option 2"]})
        >>> QuestionOptionProcessor._get_options_from_scenario(scenario, "options")
        ['Option 1', 'Option 2']


        Returns:
            list | None: List of options if found in scenario, None otherwise
        """
        scenario_options = scenario.get(option_key)
        return scenario_options if isinstance(scenario_options, list) else None

    @staticmethod
    def _get_options_from_prior_answers(
        prior_answers: dict, option_key: str
    ) -> Union[list, None]:
        """
        Try to get options from prior answers.

        prior_answers (dict): Dictionary of prior answers
        option_key (str): Key to look up in prior answers

        >>> from edsl import QuestionList as Q
        >>> q = Q.example()
        >>> q.answer = ["Option 1", "Option 2"]
        >>> prior_answers = {"options": q}
        >>> QuestionOptionProcessor._get_options_from_prior_answers(prior_answers, "options")
        ['Option 1', 'Option 2']
        >>> QuestionOptionProcessor._get_options_from_prior_answers(prior_answers, "wrong_key") is None
        True

        Returns:
            list | None: List of options if found in prior answers, None otherwise
        """
        prior_answer = prior_answers.get(option_key)
        if prior_answer and hasattr(prior_answer, "answer"):
            if isinstance(prior_answer.answer, list):
                return prior_answer.answer
        return None

    def get_question_options(self, question_data: dict) -> list:
        """
        Extract and process question options from question data.

        Args:
            question_data (dict): Dictionary containing question configuration

        Returns:
            list: List of question options. Returns default placeholders if no valid options found.

        >>> class MockPromptConstructor:
        ...     pass
        >>> mpc = MockPromptConstructor()
        >>> from edsl import Scenario
        >>> mpc.scenario = Scenario({"options": ["Option 1", "Option 2"]})
        >>> processor = QuestionOptionProcessor(mpc)

        The basic case where options are directly provided:

        >>> question_data = {"question_options": ["Option 1", "Option 2"]}
        >>> processor.get_question_options(question_data)
        ['Option 1', 'Option 2']

        The case where options are provided as a template string:

        >>> question_data = {"question_options": "{{ options }}"}
        >>> processor.get_question_options(question_data)
        ['Option 1', 'Option 2']

        The case where there is a templace string but it's in the prior answers:

        >>> class MockQuestion:
        ...     pass
        >>> q0 = MockQuestion()
        >>> q0.answer = ["Option 1", "Option 2"]
        >>> mpc.prior_answers_dict = lambda: {'q0': q0}
        >>> processor = QuestionOptionProcessor(mpc)
        >>> question_data = {"question_options": "{{ q0 }}"}
        >>> processor.get_question_options(question_data)
        ['Option 1', 'Option 2']

        The case we're no options are found:
        >>> processor.get_question_options({"question_options": "{{ poop }}"})
        ['<< Option 1 - Placeholder >>', '<< Option 2 - Placeholder >>', '<< Option 3 - Placeholder >>']

        """
        options_entry = question_data.get("question_options")

        # If not a template string, return as is or default
        if not isinstance(options_entry, str):
            return options_entry if options_entry else self._get_default_options()

        # Parse template to get variable name
        option_key = self._parse_template_variable(options_entry)

        # Try getting options from scenario
        scenario_options = self._get_options_from_scenario(
            self.prompt_constructor.scenario, option_key
        )
        if scenario_options:
            return scenario_options

        # Try getting options from prior answers
        prior_answer_options = self._get_options_from_prior_answers(
            self.prompt_constructor.prior_answers_dict(), option_key
        )
        if prior_answer_options:
            return prior_answer_options

        return self._get_default_options()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from edsl.scenarios.file_methods import FileMethods
import tempfile


class MarkdownMethods(FileMethods):
    suffix = "md"

    def view_system(self):
        import os
        import subprocess

        if os.path.exists(self.path):
            try:
                if (os_name := os.name) == "posix":
                    subprocess.run(["open", self.path], check=True)  # macOS
                elif os_name == "nt":
                    os.startfile(self.path)  # Windows
                else:
                    subprocess.run(["xdg-open", self.path], check=True)  # Linux
            except Exception as e:
                print(f"Error opening Markdown: {e}")
        else:
            print("Markdown file was not found.")

    def view_notebook(self):
        from IPython.display import FileLink, Markdown, display

        # First display the content of the markdown file
        with open(self.path, "r", encoding="utf-8") as f:
            content = f.read()
            display(Markdown(content))

        # Then provide a download link
        display(FileLink(self.path))

    def example(self):
        markdown_content = """# Sample Markdown
        
## Features
- **Bold text** demonstration
- *Italic text* demonstration
- Code block example:
```python
print("Hello, World!")
```
"""
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=".md", mode="w", encoding="utf-8"
        ) as f:
            f.write(markdown_content)
        return f.name



"""This module contains the Answers class, which is a helper class to hold the answers to a survey."""

from collections import UserDict
from edsl.data_transfer_models import EDSLResultObjectInput


class Answers(UserDict):
    """Helper class to hold the answers to a survey."""

    def add_answer(
        self, response: EDSLResultObjectInput, question: "QuestionBase"
    ) -> None:
        """Add a response to the answers dictionary."""
        answer = response.answer
        comment = response.comment
        generated_tokens = response.generated_tokens
        # record the answer
        if generated_tokens:
            self[question.question_name + "_generated_tokens"] = generated_tokens
        self[question.question_name] = answer
        if comment:
            self[question.question_name + "_comment"] = comment

    def replace_missing_answers_with_none(self, survey: "Survey") -> None:
        """Replace missing answers with None. Answers can be missing if the agent skips a question."""
        for question_name in survey.question_names:
            if question_name not in self:
                self[question_name] = None

    def to_dict(self):
        """Return a dictionary of the answers."""
        return self.data

    @classmethod
    def from_dict(cls, d):
        """Return an Answers object from a dictionary."""
        return cls(d)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from fastapi import FastAPI
from pydantic import BaseModel, create_model
from typing import Callable, Optional, Type, Dict, Any, List, Union


class SurveyToApp:
    def __init__(self, survey):
        self.survey = survey
        self.app = FastAPI()

    def parameters(self):
        return self.survey.parameters

    def create_input(self) -> Type[BaseModel]:
        """
        Creates a Pydantic model based on the survey parameters.
        Returns:
            Type[BaseModel]: A dynamically created Pydantic model class
        """
        # Get parameters from survey - now calling the method
        params = self.parameters()

        # Create field definitions dictionary
        fields: Dict[str, Any] = {}

        # Since params is a set, we'll handle each parameter directly
        # Assuming each parameter in the set has the necessary attributes
        for param in params:
            # You might need to adjust these based on the actual parameter object structure
            param_name = getattr(param, "name", str(param))
            param_type = getattr(param, "type", "string")
            is_required = getattr(param, "required", True)

            # Map survey parameter types to Python types
            type_mapping = {
                "string": str,
                "integer": int,
                "float": float,
                "boolean": bool,
                "array": List,
                # Add more type mappings as needed
            }

            # Get the Python type from mapping
            python_type = type_mapping.get(param_type, str)

            if is_required:
                fields[param_name] = (python_type, ...)
            else:
                fields[param_name] = (Optional[python_type], None)

        # Add the template variable 'name' that's used in the question text
        fields["name"] = (str, ...)

        # Create and return the Pydantic model
        model_name = f"{self.survey.__class__.__name__}Model"
        return create_model(model_name, **fields)

    def create_route(self) -> Callable:
        """
        Creates a FastAPI route handler for the survey.
        Returns:
            Callable: A route handler function
        """
        input_model = self.create_input()

        async def route_handler(input_data: input_model):
            """
            Handles the API route by processing the input data through the survey.
            Args:
                input_data: The validated input data matching the created Pydantic model
            Returns:
                dict: The processed survey results
            """
            # Convert Pydantic model to dict
            data = input_data.dict()
            print(data)
            from edsl.scenarios.Scenario import Scenario

            # Process the data through the survey
            try:
                s = Scenario(data)
                results = self.survey.by(s).run()
                return {
                    "status": "success",
                    "data": results.select("answer.*").to_scenario_list().to_dict(),
                }
            except Exception as e:
                return {"status": "error", "message": str(e)}

        return route_handler

    def add_to_app(
        self, app: FastAPI, path: str = "/survey", methods: List[str] = ["POST", "GET"]
    ):
        """
        Adds the survey route to a FastAPI application.
        Args:
            app (FastAPI): The FastAPI application instance
            path (str): The API endpoint path
            methods (List[str]): HTTP methods to support
        """
        route_handler = self.create_route()
        input_model = self.create_input()

        app.add_api_route(
            path, route_handler, methods=methods, response_model=Dict[str, Any]
        )

    def create_app(self, path: str = "/survey", methods: List[str] = ["POST", "GET"]):
        """
        Creates a FastAPI application with the survey route.
        Args:
            path (str): The API endpoint path
            methods (List[str]): HTTP methods to support
        Returns:
            FastAPI: The FastAPI application instance
        """
        app = FastAPI()
        self.add_to_app(app, path=path, methods=methods)
        return app


from edsl import QuestionFreeText, QuestionList

# q = QuestionFreeText(
#     question_name="name_gender",
#     question_text="Is this customarily a boy's name or a girl's name: {{ name}}",
# )

q = QuestionList(
    question_name="examples",
    question_text="Give me {{ num }} examples of {{ thing }}",
)

survey_app = SurveyToApp(q.to_survey())

if __name__ == "__main__":
    import uvicorn

    uvicorn.run(survey_app.create_app(path="/examples"), host="127.0.0.1", port=8000)



import time
import logging
from typing import List, TYPE_CHECKING

from edsl.results.Dataset import Dataset

if TYPE_CHECKING:
    from edsl.jobs import Jobs

    # from edsl.jobs.interviews.Interview import Interview
    # from edsl.results.Dataset import Dataset
    # from edsl.agents.AgentList import AgentList
    # from edsl.scenarios.ScenarioList import ScenarioList
    # from edsl.surveys.Survey import Survey

from edsl.jobs.FetchInvigilator import FetchInvigilator
from edsl.data.CacheEntry import CacheEntry

logger = logging.getLogger(__name__)

class JobsPrompts:
    def __init__(self, jobs: "Jobs"):
        self.interviews = jobs.interviews()
        self.agents = jobs.agents
        self.scenarios = jobs.scenarios
        self.survey = jobs.survey
        self._price_lookup = None
        self._agent_lookup = {agent: idx for idx, agent in enumerate(self.agents)}
        self._scenario_lookup = {scenario: idx for idx, scenario in enumerate(self.scenarios)}

    @property
    def price_lookup(self):
        if self._price_lookup is None:
            from edsl.coop.coop import Coop

            c = Coop()
            self._price_lookup = c.fetch_prices()
        return self._price_lookup

    def prompts(self) -> "Dataset":
        """Return a Dataset of prompts that will be used.

        >>> from edsl.jobs import Jobs
        >>> Jobs.example().prompts()
        Dataset(...)
        """
        interviews = self.interviews
        interview_indices = []
        question_names = []
        user_prompts = []
        system_prompts = []
        scenario_indices = []
        agent_indices = []
        models = []
        costs = []
        cache_keys = []
        
        for interview_index, interview in enumerate(interviews):
            logger.info(f"Processing interview {interview_index} of {len(interviews)}")
            interview_start = time.time()
            
            # Fetch invigilators timing
            invig_start = time.time()
            invigilators = [
                FetchInvigilator(interview)(question)
                for question in interview.survey.questions
            ]
            invig_end = time.time()
            logger.debug(f"Time taken to fetch invigilators: {invig_end - invig_start:.4f}s")
            
            # Process prompts timing
            prompts_start = time.time()
            for _, invigilator in enumerate(invigilators):
                # Get prompts timing
                get_prompts_start = time.time()
                prompts = invigilator.get_prompts()
                get_prompts_end = time.time()
                logger.debug(f"Time taken to get prompts: {get_prompts_end - get_prompts_start:.4f}s")
                
                user_prompt = prompts["user_prompt"]
                system_prompt = prompts["system_prompt"]
                user_prompts.append(user_prompt)
                system_prompts.append(system_prompt)
                
                # Index lookups timing
                index_start = time.time()
                agent_index = self._agent_lookup[invigilator.agent]
                agent_indices.append(agent_index)
                interview_indices.append(interview_index)
                scenario_index = self._scenario_lookup[invigilator.scenario]
                scenario_indices.append(scenario_index)
                index_end = time.time()
                logger.debug(f"Time taken for index lookups: {index_end - index_start:.4f}s")
                
                # Model and question name assignment timing
                assign_start = time.time()
                models.append(invigilator.model.model)
                question_names.append(invigilator.question.question_name)
                assign_end = time.time()
                logger.debug(f"Time taken for assignments: {assign_end - assign_start:.4f}s")

                # Cost estimation timing
                cost_start = time.time()
                prompt_cost = self.estimate_prompt_cost(
                    system_prompt=system_prompt,
                    user_prompt=user_prompt,
                    price_lookup=self.price_lookup,
                    inference_service=invigilator.model._inference_service_,
                    model=invigilator.model.model,
                )
                cost_end = time.time()
                logger.debug(f"Time taken to estimate prompt cost: {cost_end - cost_start:.4f}s")
                costs.append(prompt_cost["cost_usd"])

                # Cache key generation timing
                cache_key_gen_start = time.time()
                cache_key = CacheEntry.gen_key(
                    model=invigilator.model.model,
                    parameters=invigilator.model.parameters,
                    system_prompt=system_prompt,
                    user_prompt=user_prompt,
                    iteration=0,
                )
                cache_key_gen_end = time.time()
                cache_keys.append(cache_key)
                logger.debug(f"Time taken to generate cache key: {cache_key_gen_end - cache_key_gen_start:.4f}s")
                logger.debug("-" * 50)  # Separator between iterations

            prompts_end = time.time()
            logger.info(f"Time taken to process prompts: {prompts_end - prompts_start:.4f}s")
            
            interview_end = time.time()
            logger.info(f"Overall time taken for interview: {interview_end - interview_start:.4f}s")
            logger.info("Time breakdown:")
            logger.info(f"  Invigilators: {invig_end - invig_start:.4f}s")
            logger.info(f"  Prompts processing: {prompts_end - prompts_start:.4f}s")
            logger.info(f"  Other overhead: {(interview_end - interview_start) - ((invig_end - invig_start) + (prompts_end - prompts_start)):.4f}s")

        d = Dataset(
            [
                {"user_prompt": user_prompts},
                {"system_prompt": system_prompts},
                {"interview_index": interview_indices},
                {"question_name": question_names},
                {"scenario_index": scenario_indices},
                {"agent_index": agent_indices},
                {"model": models},
                {"estimated_cost": costs},
                {"cache_key": cache_keys},
            ]
        )
        return d

    @staticmethod
    def estimate_prompt_cost(
        system_prompt: str,
        user_prompt: str,
        price_lookup: dict,
        inference_service: str,
        model: str,
    ) -> dict:
        """Estimates the cost of a prompt. Takes piping into account."""
        import math

        def get_piping_multiplier(prompt: str):
            """Returns 2 if a prompt includes Jinja braces, and 1 otherwise."""

            if "{{" in prompt and "}}" in prompt:
                return 2
            return 1

        # Look up prices per token
        key = (inference_service, model)

        try:
            relevant_prices = price_lookup[key]

            service_input_token_price = float(
                relevant_prices["input"]["service_stated_token_price"]
            )
            service_input_token_qty = float(
                relevant_prices["input"]["service_stated_token_qty"]
            )
            input_price_per_token = service_input_token_price / service_input_token_qty

            service_output_token_price = float(
                relevant_prices["output"]["service_stated_token_price"]
            )
            service_output_token_qty = float(
                relevant_prices["output"]["service_stated_token_qty"]
            )
            output_price_per_token = (
                service_output_token_price / service_output_token_qty
            )

        except KeyError:
            # A KeyError is likely to occur if we cannot retrieve prices (the price_lookup dict is empty)
            # Use a sensible default

            import warnings

            warnings.warn(
                "Price data could not be retrieved. Using default estimates for input and output token prices. Input: $0.15 / 1M tokens; Output: $0.60 / 1M tokens"
            )
            input_price_per_token = 0.00000015  # $0.15 / 1M tokens
            output_price_per_token = 0.00000060  # $0.60 / 1M tokens

        # Compute the number of characters (double if the question involves piping)
        user_prompt_chars = len(str(user_prompt)) * get_piping_multiplier(
            str(user_prompt)
        )
        system_prompt_chars = len(str(system_prompt)) * get_piping_multiplier(
            str(system_prompt)
        )

        # Convert into tokens (1 token approx. equals 4 characters)
        input_tokens = (user_prompt_chars + system_prompt_chars) // 4

        output_tokens = math.ceil(0.75 * input_tokens)

        cost = (
            input_tokens * input_price_per_token
            + output_tokens * output_price_per_token
        )

        return {
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "cost_usd": cost,
        }

    def estimate_job_cost_from_external_prices(
        self, price_lookup: dict, iterations: int = 1
    ) -> dict:
        """
        Estimates the cost of a job according to the following assumptions:

        - 1 token = 4 characters.
        - For each prompt, output tokens = input tokens * 0.75, rounded up to the nearest integer.

        price_lookup is an external pricing dictionary.
        """

        import pandas as pd

        interviews = self.interviews
        data = []
        for interview in interviews:
            invigilators = [
                FetchInvigilator(interview)(question)
                for question in self.survey.questions
            ]
            for invigilator in invigilators:
                prompts = invigilator.get_prompts()

                # By this point, agent and scenario data has already been added to the prompts
                user_prompt = prompts["user_prompt"]
                system_prompt = prompts["system_prompt"]
                inference_service = invigilator.model._inference_service_
                model = invigilator.model.model

                prompt_cost = self.estimate_prompt_cost(
                    system_prompt=system_prompt,
                    user_prompt=user_prompt,
                    price_lookup=price_lookup,
                    inference_service=inference_service,
                    model=model,
                )

                data.append(
                    {
                        "user_prompt": user_prompt,
                        "system_prompt": system_prompt,
                        "estimated_input_tokens": prompt_cost["input_tokens"],
                        "estimated_output_tokens": prompt_cost["output_tokens"],
                        "estimated_cost_usd": prompt_cost["cost_usd"],
                        "inference_service": inference_service,
                        "model": model,
                    }
                )

        df = pd.DataFrame.from_records(data)

        df = (
            df.groupby(["inference_service", "model"])
            .agg(
                {
                    "estimated_cost_usd": "sum",
                    "estimated_input_tokens": "sum",
                    "estimated_output_tokens": "sum",
                }
            )
            .reset_index()
        )
        df["estimated_cost_usd"] = df["estimated_cost_usd"] * iterations
        df["estimated_input_tokens"] = df["estimated_input_tokens"] * iterations
        df["estimated_output_tokens"] = df["estimated_output_tokens"] * iterations

        estimated_costs_by_model = df.to_dict("records")

        estimated_total_cost = sum(
            model["estimated_cost_usd"] for model in estimated_costs_by_model
        )
        estimated_total_input_tokens = sum(
            model["estimated_input_tokens"] for model in estimated_costs_by_model
        )
        estimated_total_output_tokens = sum(
            model["estimated_output_tokens"] for model in estimated_costs_by_model
        )

        output = {
            "estimated_total_cost_usd": estimated_total_cost,
            "estimated_total_input_tokens": estimated_total_input_tokens,
            "estimated_total_output_tokens": estimated_total_output_tokens,
            "model_costs": estimated_costs_by_model,
        }

        return output

    def estimate_job_cost(self, iterations: int = 1) -> dict:
        """
        Estimates the cost of a job according to the following assumptions:

        - 1 token = 4 characters.
        - For each prompt, output tokens = input tokens * 0.75, rounded up to the nearest integer.

        Fetches prices from Coop.
        """
        return self.estimate_job_cost_from_external_prices(
            price_lookup=self.price_lookup, iterations=iterations
        )



from typing import List, Optional
from io import BytesIO
import base64
from edsl.jobs.tasks.task_status_enum import TaskStatus
from edsl.Base import RepresentationMixin


class TaskHistory(RepresentationMixin):
    def __init__(
        self,
        interviews: List["Interview"] = None,
        include_traceback: bool = False,
        max_interviews: int = 10,
        interviews_with_exceptions_only: bool = False,
    ):
        """
        The structure of a TaskHistory exception

        [Interview.exceptions, Interview.exceptions, Interview.exceptions, ...]

        >>> _ = TaskHistory.example()
        ...
        """
        self.interviews_with_exceptions_only = interviews_with_exceptions_only
        self._interviews = {}
        self.total_interviews = []
        if interviews is not None:
            for interview in interviews:
                self.add_interview(interview)

        self.include_traceback = include_traceback
        self._interviews = {
            index: interview for index, interview in enumerate(self.total_interviews)
        }
        self.max_interviews = max_interviews

        # self.total_interviews = interviews
        self.include_traceback = include_traceback

        # self._interviews = {index: i for index, i in enumerate(self.total_interviews)}
        self.max_interviews = max_interviews

    def add_interview(self, interview: "Interview"):
        """Add a single interview to the history"""
        if self.interviews_with_exceptions_only and interview.exceptions == {}:
            return

        self.total_interviews.append(interview)
        self._interviews[len(self._interviews)] = interview

    @classmethod
    def example(cls):
        """ """
        from edsl.jobs.interviews.Interview import Interview

        from edsl.jobs.Jobs import Jobs

        j = Jobs.example(throw_exception_probability=1, test_model=True)

        from edsl.config import CONFIG

        results = j.run(
            print_exceptions=False,
            skip_retry=True,
            cache=False,
            raise_validation_errors=True,
            disable_remote_cache=True,
            disable_remote_inference=True,
        )

        return cls(results.task_history.total_interviews)

    @property
    def exceptions(self):
        """
        >>> len(TaskHistory.example().exceptions)
        4
        """
        return [i.exceptions for k, i in self._interviews.items() if i.exceptions != {}]

    @property
    def unfixed_exceptions(self):
        """
        >>> len(TaskHistory.example().unfixed_exceptions)
        4
        """
        return [
            i.exceptions
            for k, i in self._interviews.items()
            if i.exceptions.num_unfixed() > 0
        ]

    @property
    def indices(self):
        return [k for k, i in self._interviews.items() if i.exceptions != {}]

    def __repr__(self):
        """Return a string representation of the TaskHistory."""
        return f"TaskHistory(interviews={self.total_interviews})."

    def to_dict(self, add_edsl_version=True):
        """Return the TaskHistory as a dictionary."""
        d = {
            "interviews": [
                i.to_dict(add_edsl_version=add_edsl_version)
                for i in self.total_interviews
            ],
            "include_traceback": self.include_traceback,
        }
        if add_edsl_version:
            from edsl import __version__

            d["edsl_version"] = __version__
            d["edsl_class_name"] = "TaskHistory"
        return d

    @classmethod
    def from_dict(cls, data: dict):
        """Create a TaskHistory from a dictionary."""
        if data is None:
            return cls([], include_traceback=False)

        from edsl.jobs.interviews.Interview import Interview

        interviews = [Interview.from_dict(i) for i in data["interviews"]]
        return cls(interviews, include_traceback=data["include_traceback"])

    @property
    def has_exceptions(self) -> bool:
        """Return True if there are any exceptions.

        >>> TaskHistory.example().has_exceptions
        True

        """
        return len(self.exceptions) > 0

    @property
    def has_unfixed_exceptions(self) -> bool:
        """Return True if there are any exceptions."""
        return len(self.unfixed_exceptions) > 0

    def show_exceptions(self, tracebacks=False):
        """Print the exceptions."""
        for index in self.indices:
            self.total_interviews[index].exceptions.print(tracebacks)

    def get_updates(self):
        """Return a list of all the updates."""
        updates = []
        for interview in self.total_interviews:
            for question_name, logs in interview.task_status_logs.items():
                updates.append(logs)
        return updates

    def print(self):
        from rich import print

        print(self.get_updates())

    def plot_completion_times(self):
        """Plot the completion times for each task."""
        import matplotlib.pyplot as plt

        updates = self.get_updates()

        elapsed = [update.max_time - update.min_time for update in updates]
        for i, update in enumerate(updates):
            if update[-1]["value"] != TaskStatus.SUCCESS:
                elapsed[i] = 0
        x = range(len(elapsed))
        y = elapsed

        plt.bar(x, y)
        plt.title("Per-interview completion times")
        plt.xlabel("Task")
        plt.ylabel("Time (seconds)")
        plt.show()

    def plotting_data(self, num_periods=100):
        updates = self.get_updates()

        min_t = min([update.min_time for update in updates])
        max_t = max([update.max_time for update in updates])
        delta_t = (max_t - min_t) / (num_periods * 1.0)
        time_periods = [min_t + delta_t * i for i in range(num_periods)]

        def counts(t):
            d = {}
            for update in updates:
                status = update.status_at_time(t)
                if status in d:
                    d[status] += 1
                else:
                    d[status] = 1
            return d

        status_counts = [counts(t) for t in time_periods]

        new_counts = []
        for status_count in status_counts:
            d = {task_status: 0 for task_status in TaskStatus}
            d.update(status_count)
            new_counts.append(d)

        return new_counts

    def plot(self, num_periods=100, get_embedded_html=False):
        """Plot the number of tasks in each state over time."""
        new_counts = self.plotting_data(num_periods)
        max_count = max([max(entry.values()) for entry in new_counts])

        rows = int(len(TaskStatus) ** 0.5) + 1
        cols = (len(TaskStatus) + rows - 1) // rows  # Ensure all plots fit

        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(rows, cols, figsize=(15, 10))
        axes = axes.flatten()  # Flatten in case of a single row/column

        for i, status in enumerate(TaskStatus):
            ax = axes[i]
            x = range(len(new_counts))
            y = [
                item.get(status, 0) for item in new_counts
            ]  # Use .get() to handle missing keys safely
            ax.plot(x, y, marker="o", linestyle="-")
            ax.set_title(status.name)
            ax.set_xlabel("Time Periods")
            ax.set_ylabel("Count")
            ax.grid(True)
            ax.set_ylim(0, max_count)

        # Hide any unused subplots
        for ax in axes[len(TaskStatus) :]:
            ax.axis("off")

        plt.tight_layout()

        if get_embedded_html:
            buffer = BytesIO()
            fig.savefig(buffer, format="png")
            plt.close(fig)
            buffer.seek(0)

            # Encode plot to base64 string
            img_data = base64.b64encode(buffer.getvalue()).decode("utf-8")
            buffer.close()
            return f'<img src="data:image/png;base64,{img_data}" alt="Plot">'
        else:
            plt.show()

    def css(self):
        from importlib import resources

        env = resources.files("edsl").joinpath("templates/error_reporting")
        css = env.joinpath("report.css").read_text()
        return css

    def javascript(self):
        from importlib import resources

        env = resources.files("edsl").joinpath("templates/error_reporting")
        js = env.joinpath("report.js").read_text()
        return js

    @property
    def exceptions_table(self) -> dict:
        """Return a dictionary of exceptions organized by type, service, model, and question name."""
        exceptions_table = {}
        for interview in self.total_interviews:
            for question_name, exceptions in interview.exceptions.items():
                for exception in exceptions:
                    key = (
                        exception.exception.__class__.__name__,  # Exception type
                        interview.model._inference_service_,    # Service
                        interview.model.model,                  # Model
                        question_name                           # Question name
                    )
                    if key not in exceptions_table:
                        exceptions_table[key] = 0
                    exceptions_table[key] += 1
        return exceptions_table
    
    @property
    def exceptions_by_type(self) -> dict:
        """Return a dictionary of exceptions tallied by type."""
        exceptions_by_type = {}
        for interview in self.total_interviews:
            for question_name, exceptions in interview.exceptions.items():
                for exception in exceptions:
                    exception_type = exception.exception.__class__.__name__
                    if exception_type in exceptions_by_type:
                        exceptions_by_type[exception_type] += 1
                    else:
                        exceptions_by_type[exception_type] = 1
        return exceptions_by_type

    @property
    def exceptions_by_service(self) -> dict:
        """Return a dictionary of exceptions tallied by service."""
        exceptions_by_service = {}
        for interview in self.total_interviews:
            service = interview.model._inference_service_
            if service not in exceptions_by_service:
                exceptions_by_service[service] = 0
            if interview.exceptions != {}:
                exceptions_by_service[service] += len(interview.exceptions)
        return exceptions_by_service

    @property
    def exceptions_by_question_name(self) -> dict:
        """Return a dictionary of exceptions tallied by question name."""
        exceptions_by_question_name = {}
        for interview in self.total_interviews:
            for question_name, exceptions in interview.exceptions.items():
                question_type = interview.survey._get_question_by_name(
                    question_name
                ).question_type
                if (question_name, question_type) not in exceptions_by_question_name:
                    exceptions_by_question_name[(question_name, question_type)] = 0
                exceptions_by_question_name[(question_name, question_type)] += len(
                    exceptions
                )

        for question in self.total_interviews[0].survey.questions:
            if (
                question.question_name,
                question.question_type,
            ) not in exceptions_by_question_name:
                exceptions_by_question_name[
                    (question.question_name, question.question_type)
                ] = 0

        sorted_exceptions_by_question_name = {
            k: v
            for k, v in sorted(
                exceptions_by_question_name.items(),
                key=lambda item: item[1],
                reverse=True,
            )
        }
        return sorted_exceptions_by_question_name

    # @property
    # def exceptions_by_model(self) -> dict:
    #     """Return a dictionary of exceptions tallied by model and question name."""
    #     exceptions_by_model = {}
    #     for interview in self.total_interviews:
    #         model = interview.model.model
    #         service = interview.model._inference_service_
    #         if (service, model) not in exceptions_by_model:
    #             exceptions_by_model[(service, model)] = 0
    #         if interview.exceptions != {}:
    #             exceptions_by_model[(service, model)] += len(interview.exceptions)

    #     # sort the exceptions by model
    #     sorted_exceptions_by_model = {
    #         k: v
    #         for k, v in sorted(
    #             exceptions_by_model.items(), key=lambda item: item[1], reverse=True
    #         )
    #     }
    #     return sorted_exceptions_by_model

    @property
    def exceptions_by_model(self) -> dict:
        """Return a dictionary of exceptions tallied by model and question name."""
        exceptions_by_model = {}
        for interview in self.total_interviews:
            model = interview.model.model
            service = interview.model._inference_service_
            for question_name, exceptions in interview.exceptions.items():
                key = (service, model, question_name)
                if key not in exceptions_by_model:
                    exceptions_by_model[key] = 0
                exceptions_by_model[key] += len(exceptions)
        return exceptions_by_model

    def generate_html_report(self, css: Optional[str], include_plot=False):
        if include_plot:
            performance_plot_html = self.plot(num_periods=100, get_embedded_html=True)
        else:
            performance_plot_html = ""

        if css is None:
            css = self.css()

        models_used = set([i.model.model for index, i in self._interviews.items()])

        from jinja2 import Environment, FileSystemLoader
        from edsl.TemplateLoader import TemplateLoader

        env = Environment(loader=TemplateLoader("edsl", "templates/error_reporting"))

        # Get current memory usage at this point

        template = env.get_template("base.html")

        # Render the template with data
        output = template.render(
            interviews=self._interviews,
            css=css,
            javascript=self.javascript(),
            num_exceptions=len(self.exceptions),
            performance_plot_html=performance_plot_html,
            exceptions_table=self.exceptions_table,
            exceptions_by_type=self.exceptions_by_type,
            exceptions_by_question_name=self.exceptions_by_question_name,
            exceptions_by_model=self.exceptions_by_model,
            exceptions_by_service=self.exceptions_by_service,
            models_used=models_used,
            max_interviews=self.max_interviews,
        )
        return output

    def html(
        self,
        filename: Optional[str] = None,
        return_link=False,
        css=None,
        cta="\nClick to open the report in a new tab\n",
        open_in_browser=False,
    ):
        """Return an HTML report."""
        from IPython.display import display, HTML
        import tempfile
        import os
        from edsl.utilities.utilities import is_notebook

        output = self.generate_html_report(css)

        # Save the rendered output to a file
        with open("output.html", "w") as f:
            f.write(output)

        if filename is None:
            current_directory = os.getcwd()
            filename = tempfile.NamedTemporaryFile(
                "w", delete=False, suffix=".html", dir=current_directory
            ).name

        with open(filename, "w") as f:
            with open(filename, "w") as f:
                f.write(output)

        if is_notebook():
            import html

            html_url = f"/files/{filename}"
            html_link = f'<a href="{html_url}" target="_blank">{cta}</a>'
            display(HTML(html_link))
            escaped_output = html.escape(output)
            iframe = f"""
            <iframe srcdoc="{ escaped_output }" style="width: 800px; height: 600px;"></iframe>
            """
            display(HTML(iframe))
        else:
            print(f"Exception report saved to {filename}")

        if open_in_browser:
            import webbrowser

            webbrowser.open(f"file://{os.path.abspath(filename)}")

        if return_link:
            return filename

    def notebook(self):
        """Create a notebook with the HTML content embedded in the first cell, then delete the cell content while keeping the output."""
        from nbformat import v4 as nbf
        from nbconvert.preprocessors import ExecutePreprocessor
        import nbformat
        import os

        # Use the existing html method to generate the HTML content
        output_html = self.generate_html_report(css=None)
        nb = nbf.new_notebook()

        # Add a code cell that renders the HTML content
        code_cell = nbf.new_code_cell(
            f"""
    from IPython.display import HTML, display
    display(HTML('''{output_html}'''))
            """
        )
        nb.cells.append(code_cell)

        # Execute the notebook
        ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
        ep.preprocess(nb, {"metadata": {"path": os.getcwd()}})

        # After execution, clear the cell's source code
        nb.cells[0].source = ""

        return nb


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Directed Acyclic Graph (DAG) class."""

from collections import UserDict
from graphlib import TopologicalSorter


class DAG(UserDict):
    """Class for creating a Directed Acyclic Graph (DAG) from a dictionary."""

    def __init__(self, data: dict):
        """Initialize the DAG class."""
        super().__init__(data)
        self.reverse_mapping = self._create_reverse_mapping()
        self.validate_no_cycles()

    def _create_reverse_mapping(self):
        """
        Create a reverse mapping of the DAG, where the keys are the children and the values are the parents.

        Example usage:

        .. code-block:: python

            data = {"a": ["b", "c"], "b": ["d"], "c": [], "d": []}
            dag = DAG(data)
            dag._create_reverse_mapping()
            {'b': {'a'}, 'c': {'a'}, 'd': {'b'}}

        """
        rev_map = {}
        for key, values in self.items():
            for value in values:
                rev_map.setdefault(value, set()).add(key)
        return rev_map

    def get_all_children(self, key):
        """Get all children of a node in the DAG."""
        children = set()

        def dfs(node):
            for child in self.reverse_mapping.get(node, []):
                if child not in children:
                    children.add(child)
                    dfs(child)

        dfs(key)
        return children

    def topologically_sorted_nodes(self):
        """
        Return a sequence of the DAG.

        Example usage:

        .. code-block:: python

            data = {"a": ["b", "c"], "b": ["d"], "c": [], "d": []}
            dag = DAG(data)
            dag.topologically_sorted_nodes() == ['c', 'd', 'b', 'a']
            True

        """
        return list(TopologicalSorter(self).static_order())

    def __add__(self, other_dag):
        """Combine two DAGs."""
        d = {}
        combined_keys = set(self.keys()).union(set(other_dag.keys()))
        for key in combined_keys:
            d[key] = self.get(key, set({})).union(other_dag.get(key, set({})))
        return DAG(d)
        # if textify:
        #     return DAG(self.textify(d))
        # else:
        #     return DAG(d)

    def remove_node(self, node: int) -> None:
        """Remove a node and all its connections from the DAG."""
        self.pop(node, None)
        for connections in self.values():
            connections.discard(node)
        # Adjust remaining nodes if necessary
        self._adjust_nodes_after_removal(node)

    def _adjust_nodes_after_removal(self, removed_node: int) -> None:
        """Adjust node indices after a node is removed."""
        new_dag = {}
        for node, connections in self.items():
            new_node = node if node < removed_node else node - 1
            new_connections = {c if c < removed_node else c - 1 for c in connections}
            new_dag[new_node] = new_connections
        self.clear()
        self.update(new_dag)

    @classmethod
    def example(cls):
        """Return an example of the `DAG`."""
        data = {"a": ["b", "c"], "b": ["d"], "c": [], "d": []}
        return cls(data)

    def detect_cycles(self):
        """
        Detect cycles in the DAG using depth-first search.

        :return: A list of cycles if any are found, otherwise an empty list.
        """
        visited = set()
        path = []
        cycles = []

        def dfs(node):
            if node in path:
                cycle = path[path.index(node) :]
                cycles.append(cycle + [node])
                return

            if node in visited:
                return

            visited.add(node)
            path.append(node)

            for child in self.get(node, []):
                dfs(child)

            path.pop()

        for node in self:
            if node not in visited:
                dfs(node)

        return cycles

    def validate_no_cycles(self):
        """
        Validate that the DAG does not contain any cycles.

        :raises ValueError: If cycles are detected in the DAG.
        """
        cycles = self.detect_cycles()
        if cycles:
            raise ValueError(f"Cycles detected in the DAG: {cycles}")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import aiohttp
import json
import requests
from typing import Any, List, Optional
from edsl.inference_services.rate_limits_cache import rate_limits

# from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.language_models import LanguageModel

from edsl.inference_services.OpenAIService import OpenAIService


class PerplexityService(OpenAIService):
    """Perplexity service class."""

    _inference_service_ = "perplexity"
    _env_key_name_ = "PERPLEXITY_API_KEY"
    _base_url_ = "https://api.perplexity.ai"
    _models_list_cache: List[str] = []
    # default perplexity parameters
    _parameters_ = {
        "temperature": 0.5,
        "max_tokens": 1000,
        "top_p": 1,
        "logprobs": False,
        "top_logprobs": 3,
    }

    @classmethod
    def available(cls) -> List[str]:
        return [
            "llama-3.1-sonar-huge-128k-online",
            "llama-3.1-sonar-large-128k-online",
            "llama-3.1-sonar-small-128k-online",
        ]

    @classmethod
    def create_model(
        cls, model_name="llama-3.1-sonar-large-128k-online", model_class_name=None
    ) -> LanguageModel:
        if model_class_name is None:
            model_class_name = cls.to_class_name(model_name)

        class LLM(LanguageModel):
            """
            Child class of LanguageModel for interacting with Perplexity models
            """

            key_sequence = cls.key_sequence
            usage_sequence = cls.usage_sequence
            input_token_name = cls.input_token_name
            output_token_name = cls.output_token_name

            _inference_service_ = cls._inference_service_
            _model_ = model_name

            _parameters_ = {
                "temperature": 0.5,
                "max_tokens": 1000,
                "top_p": 1,
                "frequency_penalty": 1,
                "presence_penalty": 0,
                # "logprobs": False, # Enable this returns 'Neither or both of logprobs and top_logprobs must be set.
                # "top_logprobs": 3,
            }

            def sync_client(self):
                return cls.sync_client()

            def async_client(self):
                return cls.async_client()

            @classmethod
            def available(cls) -> list[str]:
                return cls.sync_client().models.list()

            def get_headers(self) -> dict[str, Any]:
                client = self.sync_client()
                response = client.chat.completions.with_raw_response.create(
                    messages=[
                        {
                            "role": "user",
                            "content": "Say this is a test",
                        }
                    ],
                    model=self.model,
                )
                return dict(response.headers)

            def get_rate_limits(self) -> dict[str, Any]:
                try:
                    if "openai" in rate_limits:
                        headers = rate_limits["openai"]

                    else:
                        headers = self.get_headers()

                except Exception as e:
                    return {
                        "rpm": 10_000,
                        "tpm": 2_000_000,
                    }
                else:
                    return {
                        "rpm": int(headers["x-ratelimit-limit-requests"]),
                        "tpm": int(headers["x-ratelimit-limit-tokens"]),
                    }

            async def async_execute_model_call(
                self,
                user_prompt: str,
                system_prompt: str = "",
                files_list: Optional[List["Files"]] = None,
                invigilator: Optional[
                    "InvigilatorAI"
                ] = None,  # TBD - can eventually be used for function-calling
            ) -> dict[str, Any]:
                """Calls the OpenAI API and returns the API response."""
                if files_list:
                    encoded_image = files_list[0].base64_string
                    content = [{"type": "text", "text": user_prompt}]
                    content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{encoded_image}"
                            },
                        }
                    )
                else:
                    content = user_prompt
                client = self.async_client()

                messages = [
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": content},
                ]
                if system_prompt == "" and self.omit_system_prompt_if_empty:
                    messages = messages[1:]

                params = {
                    "model": self.model,
                    "messages": messages,
                    "temperature": self.temperature,
                    "max_tokens": self.max_tokens,
                    "top_p": self.top_p,
                    "frequency_penalty": self.frequency_penalty,
                    "presence_penalty": self.presence_penalty,
                    # "logprobs": self.logprobs,
                    # "top_logprobs": self.top_logprobs if self.logprobs else None,
                }
                try:
                    response = await client.chat.completions.create(**params)
                except Exception as e:
                    print(e, flush=True)
                return response.model_dump()

        LLM.__name__ = "LanguageModel"

        return LLM



from functools import wraps
from threading import RLock
import inspect


def synchronized_class(wrapped_class):
    """Class decorator that makes all methods thread-safe."""

    # Add a lock to the class
    setattr(wrapped_class, "_lock", RLock())

    # Get all methods from the class
    for name, method in inspect.getmembers(wrapped_class, inspect.isfunction):
        # Skip magic methods except __getitem__, __setitem__, __delitem__
        if name.startswith("__") and name not in [
            "__getitem__",
            "__setitem__",
            "__delitem__",
        ]:
            continue

        # Create synchronized version of the method
        def create_synchronized_method(method):
            @wraps(method)
            def synchronized_method(*args, **kwargs):
                instance = args[0]  # first arg is self
                with instance._lock:
                    return method(*args, **kwargs)

            return synchronized_method

        # Replace the original method with synchronized version
        setattr(wrapped_class, name, create_synchronized_method(method))

    return wrapped_class
